import operator
import math
import random
import operator as op

from functools import reduce
from threading import Thread
from collections import defaultdict
from typing import (
    Any, Dict, Type, Callable, Text, List, Set, Optional
)

import numpy as np
import pandas as pd
from py import process
import pygame as pg

from numpy import ndarray as Array
from numpy_ringbuffer import RingBuffer
from pygame.constants import KSCAN_J

from tunafish import AutomaticFunction


class GrowthStategy(AutomaticFunction):
    def evaluate(self, strategy: Callable, creature: 'Creature') -> float:
        pass

    def interface(self, health: float, energy: float) -> str:
        raise NotImplemented

    def const_strength(self) -> float:
        return self.context['creature'].strength

    def const_longevity(self) -> float:
        return self.context['creature'].longevity



class Light:
    def __init__(self, env, position, radius, intensity: int = 1):
        self.env = env
        self.position = np.array(position)
        self.radius = radius
        self.intensity = intensity
        self.is_attached_to_cursor = False

    def attach_to_cursor(self):
        self.is_attached_to_cursor = True


class Environment:
    def __init__(self, grid_shape: tuple, cell_size: tuple):
        self.grid_shape = np.array(grid_shape)
        self.cell_count = np.product(grid_shape)
        self.cell_size = np.array(cell_size)
        self.grid = np.empty(grid_shape, dtype=Cell)
        self.creatures = set()
        self.lights = []

    def add_light(self, position, radius, intensity=1.0) -> Light:
        light = Light(self, position, radius, intensity)
        self.lights.append(light)
        return light

    def add_creature(self, position: tuple) -> 'Creature':
        creature = Creature(self, None)
        creature.add_cell(position)
        self.creatures.add(creature)
        return creature

    def update(self):
        # calculate the net "energy received" by each
        # creature from all the lights
        energies_received = defaultdict(float)
        illuminated_cell_counts = defaultdict(int)
        cell_energies_received = defaultdict(float)

        for light in self.lights:

            if light.is_attached_to_cursor:
                light.position = np.array(
                    np.array(pg.mouse.get_pos()) / self.cell_size, dtype=int
                )

            # TODO: Use numpy query to select the neighborhood rather than this loop
            for i in range(-light.radius, light.radius):
                i += light.position[0]
                if i < 0 or i >= self.grid_shape[0]:
                    # skip cells out of x bounds
                    continue 
                for j in range(-light.radius, light.radius):
                    j += light.position[1]
                    if j < 0 or j >= self.grid_shape[1]:
                        # skip cells out of y bounds
                        continue

                    # get illuminated cell
                    cell = self.grid[i, j]
                    if cell is None:
                        continue

                    # compute "energy received" for the cell
                    dist = np.linalg.norm(cell.position - light.position)
                    if dist > 0:
                        energy_received = light.intensity / dist**2
                    else:
                        energy_received = light.intensity

                    # XXX
                    # if cell is not None:
                    #     print(light.position, cell.position, energy_received)

                    # increment the number of cells illuminated for
                    # the creature that owns the cell
                    illuminated_cell_counts[cell.creature] += 1

                    # increment net energy received for the creature
                    energies_received[cell.creature] += energy_received

                    # increment energy received for the individual cell
                    cell_energies_received[cell] += energy_received

        # set the specific energy received on each cell that received it
        # this is used for purposes of adding visual effects to the cell
        for cell, energy_received in cell_energies_received.items():
            cell.energy_received = energy_received

        # update the cumulative energy state of each creature
        for creature in self.creatures:
            creature.update(
                energy_received=energies_received[creature],
                energized_proportion=(
                    illuminated_cell_counts[creature] / len(creature.cells)
                )
            )


class Cell:
    def __init__(self, creature, position):
        self.position = np.array(position)
        self.creature = creature
        self.energy_received = 0.0
        self.color = np.zeros(3)
        self.rect = pg.Rect(
            creature.env.cell_size[0] * position[0],
            creature.env.cell_size[1] * position[1],
            creature.env.cell_size[0],
            creature.env.cell_size[1],
        )


class Creature:
    def __init__(self, env: Environment, strategy: GrowthStategy):
        self.env = env
        self.strategy = strategy
        self.cells: Set[Cell] = set()
        self.energies_received = RingBuffer(capacity=100)
        self.energies_transfered = RingBuffer(capacity=100)
        self.energies_usable = RingBuffer(capacity=100)
        self.longevity : float = 1.0
        self.strength: float = 0.1
        self.age: float = 0.0
        self.health = 1 / (1 + math.exp(self.longevity - self.strength * self.age))
        self.usable_energy: float = 0.0
        self.color = np.random.randint(25, 80, 3)
        self.rects = []

    def add_cell(self, position: tuple) -> Optional[Cell]:
        cell = None
        if self.env.grid[position] is None:
            cell = Cell(self, position)
            cell.energy_received = self.usable_energy
            self.cells.add(cell)
            self.env.grid[position] = cell
            self.rects.append(cell.rect)
        # returns none if cell couldn't be added
        return cell

    def update(
        self,
        energy_received: float,
        energized_proportion: float,
    ):
        # calculate the "metabolic health" of the creature in terms of its
        # efficiency in transfering received energy to a usable form.
        # efficiency near 0 means the creature is dead.
        efficiency = 1 / (1 + math.exp(self.longevity - self.strength * self.age))
        energy_transfered = efficiency * energy_received

        self.health = efficiency

        if efficiency < 0.001:
            self.usable_energy = 0
            return

        self.energies_received.append(energy_received)
        self.energies_transfered.append(energy_transfered)

        # a creature with fewer cells illuminated by a light source will
        # experience more of a lag time in absorption of energy, as energy must
        # spread out to those non-illuminated cells. Using an EMA simulates
        # this lag in energy levels.
        ema_period = round((1 - energized_proportion) * len(self.cells))
        usable_energy = pd.Series(self.energies_transfered).ewm(
            span=max(1, min(len(self.energies_transfered), ema_period))
        ).mean().iloc[-1]

        self.energies_usable.append(usable_energy)
        self.usable_energy = usable_energy
        self.age += 0.01

    def draw(self, screen):
        color = self.color.copy()
        for cell in self.cells:
            energy = cell.energy_received
            max_delta = 0xff - color[0]
            color[0] += int(max_delta * min(1, energy))
            cell.color = color
            screen.fill(cell.color, cell.rect)
        return self.rects


class Simulation:
    def __init__(self, screen_size: tuple, grid_shape: tuple, fps: int = 42):
        self.screen = None
        self.screen_size = np.array(screen_size)
        self.clock = pg.time.Clock()
        self.fps = fps
        self.running = False
        self.env = Environment(
            grid_shape=grid_shape,
            cell_size=np.round(screen_size / np.array(grid_shape))
        )
    
    def setup(self):
        if not pg.get_init():
            pg.init()
        self.screen = pg.display.set_mode(self.screen_size)

    def start(self):
        self.setup()
        self.running = True
        while self.running:
            self.clock.tick(self.fps)
            for event in pg.event.get():
                self.handle_event(event)
            self.update_game_state()
            self.update_display(self.draw())

    def update_game_state(self):
        self.env.update()

    def draw(self):
        # import ipdb; ipdb.set_trace()
        rects_to_update = []
        for creature in self.env.creatures:
            # TODO: only update changed cells
            rects_to_update.extend(creature.draw(self.screen))
        return rects_to_update

    def update_display(self, rects_to_update):
        pg.display.update(rects_to_update)

    def handle_event(self, event):
        if event.type == pg.QUIT:
            self.running = False


if __name__ == '__main__':
    sim = Simulation((1000, 1000), (100, 100))

    positions = set([tuple(np.random.randint(0, 99, 2)) for i in range(60)])
    for position in positions:
        sim.env.add_creature(position)

    light = sim.env.add_light(position=(45, 45), radius=50, intensity=50.0)
    light.attach_to_cursor()

    sim.start()


import math
import random
import operator as op

from typing import (
    Any, Dict, Type, Callable, Text
)

import numpy as np

from numpy import ndarray as Array

from tunafish import AutomaticFunction


class SymbolicRegression(AutomaticFunction):

    def stub(self, x: float) -> float:
        """
        This defines the signature of the function to evolve. After a
        SymbolicRegression instance evolves, we can access this interface via
        self.__call__, like so:

        ```python
        reg = SymbolicRegression(...)
        y = reg(x)
        ```
        """
        raise NotImplemented

    def fitness(self, regression: Callable, x: Array) -> float:
        """
        Mean squared error. We're trying to minimize it.
        """
        objective = self.context['objective']
        return np.mean(np.fromiter(
            ((regression(x_i) - objective(x_i))**2 for x_i in x), dtype=float
        ))

    def function_add(self, a: float, b: float) -> float:
        return a + b

    def function_sub(self, a: float, b: float) -> float:
        return a - b

    def function_mul(self, a: float, b: float) -> float:
        return a * b

    def function_div(self, a: float, b: float) -> float:
        return a / b if b else 1.0

    def function_neg(self, a: float) -> float:
        return -a

    def function_sin(self, a: float) -> float:
        return math.sin(a)

    def function_cos(self, a: float) -> float:
        return math.cos(a)

    def var_random_float(self) -> float:
        return float(random.randint(-1, 1))


if __name__ == '__main__':
    from matplotlib import pyplot as pp

    objective = lambda x: x**4 + x**3 + x**2 + x

    x = np.arange(-10, 10)

    regress = SymbolicRegression(
        generations=100,
        context={'objective': objective},
        max_tree_height=2,
        # NOTE: negative weight means we're minimizing the fitness value.
        # by default, we maximizing the value.
        weight=-1.0,
    ).evolve(x)

    t = np.arange(-10, 10, 0.01)
    y = np.array([objective(x_i) for x_i in t], dtype=float)
    y_hat = np.array([regress(x_i) for x_i in t], dtype=float)

    print(str(regress.population[0]))

    pp.title('Symbolic Regression')
    pp.plot(t, y, label='Objective')
    pp.plot(t, y_hat, label='Regression')
    pp.show()


import inspect

from typing import Dict, List, Callable, Optional, Union
from inspect import Parameter

import numpy as np

from .genetic_algorithm import GeneticAlgorithm
from .parameter_specification import ParameterSpecification
from .arguments import Arguments, KeywordArguments

VAR_POSITIONAL = Parameter.VAR_POSITIONAL
VAR_KEYWORD = Parameter.VAR_KEYWORD
POSITIONAL_ONLY = Parameter.POSITIONAL_ONLY


class FunctionTuner:
    """
    FunctionTuner determines which arguments to a function maximize its
    floating point return value, AKA its fitness.
    """

    def __init__(
        self,
        epochs: int = 256,
        population: Union[int, List[List[float]]] = 50,
        probabilities: Dict = GeneticAlgorithm.DEFAULT_PROBABILITIES,
        goal: Optional[float] = None,
        statistics: bool = False,
        use_multiprocessing: bool = False,
    ):
        self.specs = []
        self.epochs = max(1, epochs)
        self.goal = goal
        self.population = population or 50
        self.probabilities = probabilities
        self.are_statistics_enabled = statistics
        self.genetic_algorithm = None
        self.use_multiprocessing = use_multiprocessing
        self.winner = None

    @property
    def epoch(self) -> int:
        if self.genetic_algorithm:
            return self.genetic_algorithm.epoch
        return 0

    def tune(
        self,
        objective: Callable[..., float],
        options: Optional[Dict] = None,
    ) -> Union[KeywordArguments, Arguments, List]:
        """
        Use a genetic algorithm to determine the best input arguments to the
        objective Python function, which maximize its return value -- that is,
        it's fitness.
        """
        # build parameter specs, used in transforming geap output into
        # corresponding argument lists
        signature = inspect.signature(objective)
        params = list(signature.parameters.values())
        if len(params) == 1 and params[0].kind == VAR_KEYWORD:
            use_kwargs = True
            for k, v in options.items():
                annotation = v.get('dtype', float)
                param = Parameter(k, POSITIONAL_ONLY, annotation=annotation)
                spec = ParameterSpecification(param, options.get(k, {}))
                self.specs.append(spec)
        else:
            use_kwargs = False
            for k, param in signature.parameters.items():
                spec = ParameterSpecification(param, options.get(k, {}))
                self.specs.append(spec)

        # run the genetic algorithm, returning optimial input args.
        # memoize them in case we want to resume tuning...
        self.genetic_algorithm = GeneticAlgorithm(
            specs=self.specs,
            probabilities=self.probabilities,
            statistics=self.are_statistics_enabled,
            use_multiprocessing=self.use_multiprocessing,
            use_kwargs=use_kwargs
        )
        self.population = self.genetic_algorithm.fit(
            objective=objective,
            population=self.population,
            epochs=self.epochs,
            goal=self.goal,
        )
        # convert lists of floats (the population) into argument tuples
        if use_kwargs:
            results = [
                KeywordArguments.build(self.specs, x)
                for x in self.population
            ]
        else:
            results = [
                Arguments.build(self.specs, x)
                for x in self.population
            ]

        self.winner = results[-1]
        for x in reversed(results[:-1]):
            if x.fitness > self.winner.fitness:
                self.winner = x

        # return the first individual (an argument tuple/dict) from the final
        # evolved population; otherwise, return the entire population.
        return self.winner

    def plot(self):
        assert self.genetic_algorithm
        assert self.genetic_algorithm.are_statistics_enabled

        from matplotlib import pyplot

        stats = self.genetic_algorithm.statistics
        x = np.arange(len(stats['fitness']))
        y = stats['fitness']

        fig, ax = pyplot.subplots(1, figsize=(14,6))

        fig.suptitle(
            f'Max Fitness Per Epoch'
            f'  (Fitness Goal: {self.goal:.2f})'
                if self.goal is not None else '',
            x=0.125, y=0.98, ha='left', fontsize=18
        )
        ax.set_title(
            f'Hall of Fame:  {self.winner if self.winner else ""}',
            loc='left', fontsize=10, fontname='monospace',
        )
        ax.set_xlabel('Epoch')
        ax.set_ylabel('Fitness')
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.yaxis.grid(color='gray', linestyle='dashed', alpha=0.3)
        ax.xaxis.grid(color='gray', linestyle='dashed', alpha=0.3)
        
        ax.plot(x, y, linewidth=2.0)
        pyplot.show()


import sys
import random
import os

from multiprocessing import Pool
from datetime import datetime
from typing import (
    Dict, List, Callable, Optional, Union, Sequence, Tuple
)

import deap.base
import deap.creator
import deap.tools
import numpy as np

from .parameter_specification import ParameterSpecification
from .arguments import Arguments


class GeneticAlgorithm:
    """
    This GeneticAlgorithm class is just a convenient encapsulation of the
    more awkward Geap evolutionary algorithm framework. The algorithm simply
    generates populations of lists of floats. Each float gets
    fitted/translated into a corresponding argument value expected by the
    Python function being optimized.
    """

    DEFAULT_PROBABILITIES: Dict = {
        'crossover': 0.5,
        'mutation': 0.25,
    }

    def __init__(
        self,
        specs: List[ParameterSpecification],
        probabilities: Dict = DEFAULT_PROBABILITIES,
        statistics: bool = False,
        use_multiprocessing: bool = False,
        use_kwargs: bool = False,
    ):
        self.specs = specs
        self.keys = tuple(spec.name for spec in self.specs)
        self.probabilities = probabilities
        self.objective: Optional[Callable] = None
        self.are_statistics_enabled = statistics
        self.statistics = None
        self.use_kwargs = use_kwargs
        self.use_multiprocessing = use_multiprocessing
        self.pool = None
        self.epoch = 0

    def setup(self):
        """
        Initialize and register types and functions used internally by the
        Deap evolutionary computing framework.
        """
        toolbox = deap.base.Toolbox()

        # define geap internal sub-classes
        if not hasattr(deap.creator, 'FunctionTunerFitness'):
            deap.creator.create(
                'FunctionTunerFitness', deap.base.Fitness, weights=(1.0, )
            )
        if not hasattr(deap.creator, 'FunctionTunerIndividual'):
            deap.creator.create(
                'FunctionTunerIndividual', list,
                fitness=deap.creator.FunctionTunerFitness
            )

        # register our function that initializes each individual
        toolbox.register(
            'initializer', self.initializer
        )
        # tell Geap how to generate an individual with our initializer
        toolbox.register(
            'individual', deap.tools.initIterate,
            deap.creator.FunctionTunerIndividual, toolbox.initializer,
        )
        # create an initial population using the registered "individual" func
        toolbox.register(
            'population',
            deap.tools.initRepeat, list, toolbox.individual
        )
        # tell geap how to select the best performing individuals each epoch
        toolbox.register(
            'select', deap.tools.selTournament, tournsize=3
        )
        # tell geap how we want to perform genetic crossover
        toolbox.register(
            'mate', deap.tools.cxTwoPoint
        )
        # set the mutation operation
        toolbox.register(
            'mutate', deap.tools.mutGaussian, mu=0, sigma=1, indpb=0.1
        )
        # set our fitness function
        toolbox.register(
            'evaluate', self.evaluate
        )

        if self.use_multiprocessing:
            pool = Pool(processes=max(1, os.cpu_count() or 1))
            toolbox.register('map', pool.map)

        return toolbox

    def fit(
        self,
        objective: Callable[..., float],
        population: Union[int, Sequence] = 25,
        epochs: int = 50,
        goal: float = None,
    ) -> List[List]:
        """
        Generate argument list (or lists) which maximize the fitness (a
        float) of the objective function. The objective function's output is
        expected to be its fitness.
        """
        toolbox = self.setup()

        self.objective = objective
        self.epoch = 0

        # initialize a new statistics object
        if self.are_statistics_enabled:
            self.statistics = {
                'fitness': [],
                'winners': [],
                'start': datetime.now(),
                'stop': None,  # : datetime
                'time': None,  # : timedelta
            }

        # generate initial population
        if isinstance(population, int):
            pop = toolbox.population(n=population)
        else:
            assert isinstance(population, (list, tuple, set))
            pop = list(population)

        is_goal_reached = False # only matters if goal is not None

        # evolve the initial population....
        for epoch in range(epochs):
            self.epoch = epoch
            print(f'Epoch {epoch}...') # TODO: replace with custom callback
            # generate next population from previous
            # using tournament selection
            offspring = list(
                map(
                    toolbox.clone,
                    toolbox.select(pop, len(pop))
                )
            )
            # apply crossovers
            for siblings in zip(offspring[::2], offspring[1::2]):
                if random.random() < self.probabilities['crossover']:
                    toolbox.mate(*siblings)
                    for child in siblings:
                        del child.fitness.values

            # apply mutations
            for child in offspring:
                if random.random() < self.probabilities['mutation']:
                    toolbox.mutate(child)
                    del child.fitness.values

            # evaluate the fitness of each individual in need of it
            invalid_offspring = [x for x in offspring if not x.fitness.valid]
            fitness_values = toolbox.map(toolbox.evaluate, invalid_offspring)
            for child, fitness in zip(invalid_offspring, fitness_values):
                child.fitness.values = fitness
                if goal is not None and fitness[0] >= goal:
                    is_goal_reached = True

            for child in offspring:
                if goal is not None and child.fitness.values[0] >= goal:
                    is_goal_reached = True

            # replace last generation with the next
            pop = offspring

            # update per-epoch statistics
            if self.are_statistics_enabled:
                # keep track of each epoch's max fitness
                max_fitness = -1 * sys.maxsize
                winner = None
                for child in offspring:
                    fitness = child.fitness.values[0]
                    if fitness > max_fitness:
                        max_fitness = fitness
                        winner = child
                self.statistics['fitness'].append(max_fitness)
                self.statistics['winners'].append(winner)

            if is_goal_reached:
                break

        # update statistics
        if self.are_statistics_enabled:
            self.statistics['fitness'] = np.array(self.statistics['fitness'])
            self.statistics['winners'] = np.array(self.statistics['winners'])
            self.statistics['stop'] = datetime.now()
            self.statistics['time'] = (
                self.statistics['stop'] - self.statistics['start']
            )

        return pop

    def initializer(self) -> List:
        """
        This function generates a new individual for our genetic algorithm.
        This function is called internally by Geap.
        """
        arg_list = []  # <- returned 
        for spec in self.specs:
            if spec.sequence_dtype is not None:
                raise NotImplementedError(
                    'TODO: use a separate GA for this list argument'
                )
            else:
                # x is our value to append to values list
                x = random.random() * 100.0  # default float value

                # compute x (override its current default value)
                if spec.enum_values:
                    x = random.randrange(len(spec.enum_values))
                elif spec.dtype is bool:
                    x = random.random()
                else:
                    x_min = spec.min_value
                    x_max = spec.max_value
                    if spec.dtype is int:
                        x = int(x_min + random.random() * (x_max - x_min))
                    if spec.dtype is float:
                        x = x_min + random.random() * (x_max - x_min)

                arg_list.append(x)

        return arg_list

    def evaluate(self, individual: List) -> Tuple[float]:
        """
        Translate genetic algorithm individual into arguments with which to
        call the objective function (that we're optimizing), call it, and
        return its fitness. This function is called internally by Geap.
        """
        args = Arguments.build(self.specs, individual)

        if self.use_kwargs:
            kwargs = dict(zip(self.keys, args))
            args.fitness = self.objective(**kwargs) or 0.0
        else:
            args.fitness = self.objective(*args) or 0.0

        return (args.fitness, )


"""
class AutomaticFunction
"""

import operator
import inspect
import pickle

from multiprocessing import Pool
from inspect import Signature, signature
from typing import Dict, Callable, Text, Tuple, Optional, Any

import numpy as np
import deap.base
import deap.creator
import deap.tools
import deap.gp
import deap.algorithms

from deap.gp import PrimitiveSetTyped

from tunafish.evolution.util import (
    clamp, get_required_parameters, is_parameter_required
)

# global args and kwargs are set by pool worker processes in their initializers.
# this only happens when AutomaticFunction is instantiated with
# multiprocessing=True.
# ---------------------------------------
GLOBAL_KWARGS = {}
GLOBAL_ARGS = tuple()


class AutomaticFunction:
    """
    An AutomaticFunction uses genetic programming to program itself according to
    which inputs optimize its output, according to a fitness function.
    """

    def __init__(
        self,
        generations: int = 64,
        population: int = 256,
        min_tree_height: int = 1,
        max_tree_height: int = 5,
        tournament_size: int = 3,
        crossover_rate: float = 0.5,
        mutation_rate: float = 0.1,
        weights: Optional[Tuple] = (1.0, ),
        weight: Optional[float] = None,
        verbose: bool = True,
        multiprocessing: bool = True,
        context: Dict = None,
        meta: dict = None,
    ):
        self.class_name = type(self).__name__
        self.n_generations = clamp(generations, 2, 100000)
        self.n_population = clamp(population, 2, 5000)
        self.min_tree_height = clamp(min_tree_height, 1, 17)
        self.max_tree_height = clamp(max_tree_height, min_tree_height, 17)
        self.tournament_size = clamp(tournament_size, 1, population // 2)
        self.crossover_rate = clamp(crossover_rate, 0.01, 1.0)
        self.mutation_rate = clamp(mutation_rate, 0.01, 1.0)
        self.verbose = verbose
        self.context = context or {}
        self.meta = meta or {}
        self.primitives = None
        self.registered_names = set()
        self.use_multiprocessing = multiprocessing
        self.deferred_primitives = {
            'variables': [],
            'constants': [],
            'functions': [],
        }

        self.individual_class_name = f'{self.class_name}Individual'
        self.fitness_class_name = f'{self.class_name}Fitness'

        # vars set by self.evolve:
        self.logbook = None
        self.population = None
        self.expressions = None
        self.winner = None
        self.winners = []
        self.winner_expression = None
        self.winner_expressions = []

        if weight is not None:
            self.weights = (weight, )
        else:
            assert weights
            if isinstance(weights, (list, np.ndarray)):
                self.weights = tuple(weights)
            else:
                self.weights = weights

        self.unfit = tuple(
            -np.inf if w >= 0 else np.inf
            for w in self.weights
        )

        # input_items are (arg_name, dtype) tuples...
        if isinstance(self.stub, Signature):
            sig = self.stub
        else:
            sig = inspect.signature(self.stub)

        input_items = tuple(
            (k, v.annotation)
            for k, v in list(sig.parameters.items())
            if is_parameter_required(v) and k not in ('self', 'cls')
        )
        # create container for functions and terminals used by automatic_function
        self.primitives = PrimitiveSetTyped(
            self.class_name, [item[1] for item in input_items],
            sig.return_annotation
        )
        # rename input arguments from geap defaults to
        # the keys in the self.inputs dict.
        self.primitives.renameArguments(**{
            f'ARG{i}': item[0] for i, item in enumerate(input_items)
        })

        # register Deap internal classes
        if not hasattr(deap.creator, self.fitness_class_name):
            deap.creator.create(
                f'{self.class_name}Fitness', deap.base.Fitness,
                weights=self.weights,
            )

        if not hasattr(deap.creator, self.individual_class_name):
            deap.creator.create(
                self.individual_class_name, deap.gp.PrimitiveTree,
                fitness=getattr(deap.creator, self.fitness_class_name),
            )

    def __call__(self, *args, **kwargs):
        return self.winner(*args, *kwargs) if self.winner else None

    def register_function(self, func: Callable, name: str = None) -> bool:
        name = (name or func.__name__).upper()
        if name in self.registered_names:
            return False

        if name.startswith('function_'):
            name = name[9:]

        if self.primitives is not None:
            sig = inspect.signature(func)
            params = get_required_parameters(sig)
            in_types = [p.annotation for p in params.values()]
            ret_type = sig.return_annotation
            self.primitives.addPrimitive(
                func, in_types=in_types, ret_type=ret_type, name=name
            )
        else:
            self.deferred_primitives['functions'].append((func, name))

        self.registered_names.add(name)
        return True

    def register_constant(self, func: Callable, name: str = None) -> bool:
        name = name or func.__name__
        if name in self.registered_names:
            return False
        if name.startswith('const_'):
            name = name[6:]

        if self.primitives is not None:
            value = func()
            self.primitives.addTerminal(value, type(value), name=name)
        else:
            self.deferred_primitives['constants'].append((func, name))

        self.registered_names.add(name)
        return True

    def register_variable(self, func: Callable, name: str = None) -> bool:
        name = name or func.__name__
        if name in self.registered_names:
            return False
        if name.startswith('var_'):
            name = name[4:]

        if self.primitives is not None:
            sig = signature(func)
            ret_type = sig.return_annotation
            self.primitives.addEphemeralConstant(name, func, ret_type)
        else:
            self.deferred_primitives['variables'].append((func, name))

        self.registered_names.add(name)
        return True

    def stub(self, x: float) -> float:
        raise NotImplementedError()

    def setup(self, args=None, kwargs=None):
        # add self.functions to primitives
        for key, val in inspect.getmembers(self, callable):
            prefix = key.split('_')[0].lower()
            if prefix in ('func', 'function'):
                self.register_function(val)
            elif prefix == ('var', 'variable'):
                self.register_variable(val)
            elif prefix in ('const', 'constant'):
                self.register_constant(val)

        for args in self.deferred_primitives['functions']:
            self.register_function(*args)
        for args in self.deferred_primitives['constants']:
            self.register_constant(*args)
        for args in self.deferred_primitives['variables']:
            self.register_variable(*args)

        toolbox = deap.base.Toolbox()
        # NOTE: toolbox must not be set on or stored in self, as it will cause
        # problems internally to Deap in regards to pickling and
        # multiprocessing.

        toolbox.register('expr',
            deap.gp.genHalfAndHalf, pset=self.primitives,
            min_=self.min_tree_height, max_=self.max_tree_height
        )
        toolbox.register('individual',
            deap.tools.initIterate,
            getattr(deap.creator, self.individual_class_name),
            toolbox.expr  # pylint: disable=no-member
        )
        toolbox.register(
            'population', deap.tools.initRepeat, list,
            toolbox.individual  # pylint: disable=no-member
        )
        toolbox.register(
            'compile', deap.gp.compile, pset=self.primitives
        )
        toolbox.register(
            'select', deap.tools.selTournament,
            tournsize=self.tournament_size
        )
        toolbox.register(
            'mate', deap.gp.cxOnePoint
        )
        toolbox.register(
            'expr_mut', deap.gp.genFull, min_=0, max_=self.max_tree_height
        )
        toolbox.register(
            'mutate', deap.gp.mutUniform,
            expr=toolbox.expr_mut,  # pylint: disable=no-member
            pset=self.primitives
        )
        # for increased parallelism and speed...
        if self.use_multiprocessing:
            pool = Pool(
                processes=8,
                initializer=self.process_initializer,
                initargs=(args, kwargs)
            )
            toolbox.register('map', pool.map)
            toolbox.register(
                'evaluate', self.compile_and_evaluate,
            )
        else:
            toolbox.register(
                'evaluate', self.compile_and_evaluate, args=args, kwargs=kwargs
            )

        # set hard limit for total tree size to 17, as recommended by Koza
        toolbox.decorate(
            'mate', deap.gp.staticLimit(
                key=operator.attrgetter('height'), max_value=17
            )
        )
        toolbox.decorate(
            'mutate', deap.gp.staticLimit(
                key=operator.attrgetter('height'), max_value=17
            )
        )

        return toolbox

    @staticmethod
    def process_initializer(args: tuple, kwargs: Dict):
        """
        This is the initializer method used by each worker process managed by
        the process Pool. This method is responsible to sharing arguments to the
        fitness function between processes without having to share them via
        pickling, which is what Deap does otherwise, which can be very slow.
        """
        global GLOBAL_ARGS     # pylint: disable=global-statement
        global GLOBAL_KWARGS   # pylint: disable=global-statement

        GLOBAL_ARGS = args
        GLOBAL_KWARGS = kwargs

    def evolve(self, *args, **kwargs) -> 'AutomaticFunction':
        """
        Train the AutomaticFunction by running genetic programming.
        """
        toolbox = self.setup(args, kwargs)

        # configure statistics for evolve loop
        stats_fit = deap.tools.Statistics(lambda ind: ind.fitness.values)
        stats_size = deap.tools.Statistics(len)
        stats = deap.tools.MultiStatistics(
            fitness=stats_fit, size=stats_size
        )
        stats.register('avg', np.nanmean)
        stats.register('std', np.nanstd)
        stats.register('min', np.nanmin)
        stats.register('max', np.nanmax)

        # initialize generation 0
        pop = toolbox.population(  # pylint: disable=no-member
            n=self.n_population
        ) 
        # kick off the evolve loop...
        hof = deap.tools.HallOfFame(10)
        final_pop, self.logbook = deap.algorithms.eaSimple(
            pop, toolbox, self.crossover_rate, self.mutation_rate,
            self.n_generations, stats=stats, halloffame=hof,
            verbose=self.verbose
        )

        # memoize the final population in the form of compile funcs
        self.population = final_pop
        self.expressions = [
            deap.gp.compile(individual, self.primitives)
            for individual in final_pop
        ]
        # set the most fit individual (evolved function)
        self.winner = deap.gp.compile(hof[-1], self.primitives)
        self.winner_expression = hof[-1]
        self.winner_expressions = list(hof)
        self.winners = [
            deap.gp.compile(expr, self.primitives)
            for expr in hof
        ]

        return self

    def compile_and_evaluate(self, individual, args=None, kwargs=None) -> Tuple:
        evolved_func = deap.gp.compile(individual, self.primitives)
        args = args or GLOBAL_ARGS
        kwargs = kwargs or GLOBAL_KWARGS
        fitness = self.fitness(evolved_func, *args, **kwargs)
        if isinstance(fitness, (int, float, np.number)):
            return (float(fitness), )
        elif isinstance(fitness, (list, np.ndarray)):
            return tuple(fitness)
        else:
            assert isinstance(fitness, tuple)
            return fitness

    def compile(self, individual) -> Callable:
        """
        Public interface method for compiling a Deap individual into a lambda
        function. This can be useful when exploring self.winner_expressions,
        etc. at the end of an evolution/training.
        """
        return deap.gp.compile(individual, self.primitives)

    @property
    def source(self) -> Optional[str]:
        """
        Return source code string of the evolution winner.
        """
        return (
            str(self.winner_expression) if self.winner_expression else None
        )

    def fitness(self, individual: Callable, *args, **kwargs):
        """
        Returns any of the following:
        float, int, list, tuple, np.ndarray, np.number
        """
        raise NotImplementedError()

    def pickle(self, path: str, meta: dict = None) -> dict:
        """
        Persist the AutomaticFunction's internal state as a pickled dict.
        """
        if meta:
            self.meta.update(meta)
        data = {
            # NOTE: we must store some fields as pickles inside the main pickle
            # in order to accomodate limitations in unpickling Deap components.
            'population': pickle.dumps(self.population),
            'winners': pickle.dumps(self.winner_expressions),

            'logbook': self.logbook,
            'kwargs': {
                'generations': self.n_generations,
                'population': self.n_population,
                'min_tree_height': self.min_tree_height,
                'max_tree_height': self.max_tree_height,
                'tournament_size': self.tournament_size,
                'crossover_rate': self.crossover_rate,
                'mutation_rate': self.mutation_rate,
                'weights': self.weights,
                'verbose': self.verbose,
                'meta': self.meta,
            }
        }
        with open(path, 'wb') as pickle_file:
            pickle.dump(data, pickle_file)
            return data

    @classmethod
    def unpickle(cls, path: str, *args, **kwargs) -> 'AutomaticFunction':
        """
        Unpickle a persisted AutomaticFunction state dict. See pickle method.
        """
        with open(path, 'rb') as pickle_file:
            state = pickle.load(pickle_file)
            ctor_kwargs = state['kwargs']

        func = cls(**ctor_kwargs)
        func.setup(args=args, kwargs=kwargs)

        # func.setup must be called before Deap components are unpickled. This
        # forces us to store the "population" as a pickle inside the pickle
        # itself -- so that we can unpickled it after calling setup.
        population = pickle.loads(state['population'])
        winner_expressions = pickle.loads(state['winners'])

        expressions = [func.compile(x) for x in population]
        winners = [func.compile(x) for x in winner_expressions]

        func.logbook = state['logbook']
        func.population = population
        func.expressions = expressions
        func.winner_expressions = winner_expressions
        func.winner_expression = winner_expressions[-1]
        func.winners = winners
        func.winner = winners[-1]

        return func



from typing import Dict

from inspect import Parameter, Signature


def is_parameter_required(parameter: Parameter) -> bool:
    return parameter.kind in (
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.POSITIONAL_ONLY
    )


def get_required_parameters(signature: Signature) -> Dict:
    return {
        k: v for k, v in signature.parameters.items()
        if is_parameter_required(v)
    }


def clamp(x, lower=0.0, upper=1.0):
    # XXX: Deprecated. Use np.clip instead
    return min(max(lower, x), upper)


from tunafish import FunctionTuner


def objective(name: str, age: int, temperature: float, male: bool):
    """
    This is the "objective" function whose output we're optimizing with
    respect to the inputs. In other words, we're finding the inputs which
    maximize fitness (the output of the function).
    """
    fitness = 0

    # consider name and male args as binary right/wrong
    fitness += int(name == 'Jeff')
    fitness += int(male)

    # penalize the more age and temp differs from 37 and 98.6
    fitness -= abs(age - 37)
    fitness -= abs(temperature - 98.6)

    return fitness
        

if __name__ == '__main__':
    # optional constraints on the argument search space
    options = {
        'age': {'min': 30, 'max': 50},
        'temperature': {'min': 91.0, 'max': 104},
        'name': {'enum': ['Jeff', 'John', 'Jimbob', 'Joffrey']},
    }

    # maximum of 500 training epochs
    epochs = 500

    # break early if any individual reaches fitness >= 1.99
    goal = 1.99

    tuner = FunctionTuner(epochs=epochs, goal=goal, statistics=True)
    tuner.tune(objective, options)
    tuner.plot()


from typing import List, Optional


class Arguments(tuple):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.fitness: float = 0
        self.individual: Optional[List[float]] = None

    @classmethod
    def build(cls, specs, individual: List[float]) -> 'Arguments':
        """
        Convert raw genetic algorithm output floats into function argument
        values.
        """
        return cls([s.fit(x) for s, x in zip(specs, individual)])


class KeywordArguments(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fitness: float = 0
        self.individual: Optional[List[float]] = None

    @classmethod
    def build(cls, specs, individual: List[float]) -> 'KeywordArguments':
        """
        Convert raw genetic algorithm output floats into function argument
        values.
        """
        return cls({s.name: s.fit(x) for s, x in zip(specs, individual)})


from inspect import Parameter
from typing import (
    Dict, Any, Type, Tuple, List, Text,
    Set, Optional
)


class ParameterSpecification:
    """
    The ParameterSpecification contains data necessary to convert the output
    of a genetic algorithm into an int, float, string, bool or other datatype
    expected by the function who owns the parameter.

    The genetic algorithm generates a float for each argument expected by the
    parameter's owner function. Each float must be converted into a
    corresponding int, float, bool, string or other datatype expected by the
    function.

    Note that, for each non-scalar parameter -- that is, for lists, tuple,
    and sets -- the sequence is generated by a separate genetic algorithm,
    forming a tree of genetic algorithms.
    """

    def __init__(self, param: Parameter, options: Dict):
        self.param = param
        self.options = options
        self.dtype, self.sequence_dtype = self._infer_type(param)

        # NOTE: sequence_dtype is non-None if this spec is for an annotation
        # like `List[int]` or `tuple`. For `List[int]`, self.dtype is int, and
        # self.sequence_dtype is list.

        # enum_values is a list of the parameter's allowed values
        self.enum_values: List = self.options.get('enum', [])

        # min and max numerical values
        self.min_value: Any[int, float] = self.options.get('min', -1.0)
        self.max_value: Any[int, float] = self.options.get('max', 1.0)

        # select the proper "fit" function based on the parameter's datatype.
        # The fit function converts a raw float, generated by a genetic
        # algorithm, into a corresponding argument value.
        if self.dtype is int:
            self._fit_func = self._fit_int
            # assume we generally want positive int values, like list indices,
            # so set min and max different from default
            if self.min_value is None:
                self.min_value = 0
            if self.max_value is None:
                self.max_value = 1000
        elif self.dtype is str:
            self._fit_func = self._fit_string
            self.max_value = None
            self.min_value = None
        elif self.dtype is bool:
            self._fit_func = self._fit_bool
        else:
            self._fit_func = self._fit_float

        # if the parameter is non-scalar (e.g. is a list, set, tuple),
        # self.size defines its desired length.
        if self.sequence_dtype is not None:
            self.size = self.options.get('size')
            if self.size is None or self.size < 1:
                raise ValueError(f'{param} options must specify a size')
        else:
            self.size = 1

    @property
    def name(self) -> str:
        return self.param.name

    def fit(self, x: float) -> object:
        """
        Convert a raw floating point value, generated by a genetic algorithm,
        into a corresponding parameter value to apply to the owner function
        (which we are optimizing).
        """
        return self._fit_func(x)  # delegate based on the self.dtype

    def _fit_float(self, x: float) -> float:
        """
        Return valid float argument from raw genetic algorithm output value.
        """
        if self.min_value is not None:
            x = max(self.min_value, x)
        if self.max_value is not None:
            x = min(self.max_value, x)
        return x

    def _fit_int(self, x: float) -> int:
        """
        Return valid int argument from raw genetic algorithm output value.
        """
        if self.min_value is not None:
            x = max(self.min_value, x)
        if self.max_value is not None:
            x = min(self.max_value, x)
        return int(x)

    def _fit_string(self, x: float) -> str:
        """
        Return valid str argument from raw genetic algorithm output value.
        """
        values = self.enum_values
        return values[int(abs(x)) % len(values)]

    def _fit_bool(self, x: float) -> bool:
        """
        Return valid bool argument from raw genetic algorithm output value.
        """
        return x >= 0

    def _infer_type(self, param: Parameter) -> Tuple[Type, Optional[Type]]:
        """
        Givent an inspect.Parameter's annotation, resolve and return
        corresponding type object(s). If the annotation is a scalar type,
        like an int or str, we return just (int, None); however, if the
        annotation is something like `List[int]`, we'd return (int, list). By
        default, we return (float, None) -- that is, a single float dtype.
        """
        scalar_types = (int, str, float, bool)
        default_dtype = float
        ann = param.annotation

        if not ann:
            # use defaults
            return (default_dtype, None)

        if isinstance(ann, type):
            if issubclass(ann, scalar_types):
                # the parameter is an int, float, str, etc.
                return (ann, None)
            elif issubclass(ann, (tuple, list, set)):
                # we have a sequence but non-specified inner type,
                # so we use the default (float)
                return (default_dtype, ann)
        elif ann is Text:
            # Text is an alias for str
            return (ann, None)
        elif ann in (Tuple, List, Set):
            # extract inner type from non-scalar annotation, like List[int]
            dtype_map = {Tuple: tuple, List: list, Set: set}
            seq_dtype = dtype_map[ann]
            if ann.__args__:
                if len(ann.__args__) > 1:
                    raise ValueError(
                        f'cannot autotune sequences of anything'
                        f'but floats and ints but got {param}'
                    )
                inner_dtype  = ann.__args__[0]
                if inner_dtype in scalar_types or inner_dtype is Text:
                    return (inner_dtype, seq_dtype)
            else:
                # use defaults
                return (default_dtype, seq_dtype)

        raise ValueError(f'cannot autotune {param} parameter')
